import { Backend, DatabaseConnection, DatabaseTransaction, Schema, RecordStoreRequest, RecordGetRequest, RecordGetResponse, RecordStoreResponse } from "./backend-interface";
import { Key, Value } from "./util/types";
import BridgeIDBKeyRange from "./BridgeIDBKeyRange";
interface IndexDump {
    name: string;
    records: IndexRecord[];
}
interface ObjectStoreDump {
    name: string;
    keyGenerator: number;
    records: ObjectStoreRecord[];
    indexes: {
        [name: string]: IndexDump;
    };
}
interface DatabaseDump {
    schema: Schema;
    objectStores: {
        [name: string]: ObjectStoreDump;
    };
}
interface MemoryBackendDump {
    databases: {
        [name: string]: DatabaseDump;
    };
}
interface IndexRecord {
    indexKey: Key;
    primaryKeys: Key[];
}
interface ObjectStoreRecord {
    primaryKey: Key;
    value: Value;
}
/**
 * Primitive in-memory backend.
 */
export declare class MemoryBackend implements Backend {
    private databases;
    private connectionIdCounter;
    private transactionIdCounter;
    /**
     * Connections by connection cookie.
     */
    private connections;
    /**
     * Connections by transaction (!!) cookie.  In this implementation,
     * at most one transaction can run at the same time per connection.
     */
    private connectionsByTransaction;
    /**
     * Condition that is triggered whenever a client disconnects.
     */
    private disconnectCond;
    /**
     * Conditation that is triggered whenever a transaction finishes.
     */
    private transactionDoneCond;
    afterCommitCallback?: () => Promise<void>;
    enableTracing: boolean;
    /**
     * Load the data in this IndexedDB backend from a dump in JSON format.
     *
     * Must be called before any connections to the database backend have
     * been made.
     */
    importDump(data: any): void;
    private makeObjectStoreMap;
    /**
     * Export the contents of the database to JSON.
     *
     * Only exports data that has been committed.
     */
    exportDump(): MemoryBackendDump;
    getDatabases(): Promise<{
        name: string;
        version: number;
    }[]>;
    deleteDatabase(tx: DatabaseTransaction, name: string): Promise<void>;
    connectDatabase(name: string): Promise<DatabaseConnection>;
    beginTransaction(conn: DatabaseConnection, objectStores: string[], mode: import("./util/types").TransactionMode): Promise<DatabaseTransaction>;
    enterVersionChange(conn: DatabaseConnection, newVersion: number): Promise<DatabaseTransaction>;
    close(conn: DatabaseConnection): Promise<void>;
    getSchema(dbConn: DatabaseConnection): Schema;
    renameIndex(btx: DatabaseTransaction, objectStoreName: string, oldName: string, newName: string): void;
    deleteIndex(btx: DatabaseTransaction, objectStoreName: string, indexName: string): void;
    deleteObjectStore(btx: DatabaseTransaction, name: string): void;
    renameObjectStore(btx: DatabaseTransaction, oldName: string, newName: string): void;
    createObjectStore(btx: DatabaseTransaction, name: string, keyPath: string | string[] | null, autoIncrement: boolean): void;
    createIndex(btx: DatabaseTransaction, indexName: string, objectStoreName: string, keyPath: import("./util/types").KeyPath, multiEntry: boolean, unique: boolean): void;
    deleteRecord(btx: DatabaseTransaction, objectStoreName: string, range: BridgeIDBKeyRange): Promise<void>;
    private deleteFromIndex;
    getRecords(btx: DatabaseTransaction, req: RecordGetRequest): Promise<RecordGetResponse>;
    storeRecord(btx: DatabaseTransaction, storeReq: RecordStoreRequest): Promise<RecordStoreResponse>;
    private insertIntoIndex;
    rollback(btx: DatabaseTransaction): Promise<void>;
    commit(btx: DatabaseTransaction): Promise<void>;
}
export default MemoryBackend;
